<?php

namespace Drupal\monarch_fast_loader;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\SelectInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldStorageDefinitionInterface;

/**
 * Loads large numbers of entities quickly.
 */
final class EntityFastLoader {

  /**
   * The column info cache.
   */
  protected array $columnInfo = [];

  /**
   * Constructs an EntityFastLoader object.
   */
  public function __construct(
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly EntityFieldManagerInterface $entityFieldManager,
    private readonly Connection $database,
  ) {}

  /**
   * Gets the service.
   */
  public static function service() : self {
    return \Drupal::service('entity.fast_loader');
  }

  /**
   * Checks whether the given field can be stored in a shared table.
   */
  public function allowsSharedTableStorage(FieldStorageDefinitionInterface $storage_definition) : bool {
    return !$storage_definition->hasCustomStorage() && $storage_definition->isBaseField() && !$storage_definition->isMultiple() && !$storage_definition->isDeleted();
  }

  /**
   * Checks whether the given field has to be stored in a dedicated table.
   */
  public function requiresDedicatedTableStorage(FieldStorageDefinitionInterface $storage_definition) : bool {
    return !$storage_definition->hasCustomStorage() && !$this->allowsSharedTableStorage($storage_definition);
  }

  /**
   * Get a field column name.
   */
  public function getColumnsInfo(FieldStorageDefinitionInterface $field_storage_definition) {
    $fsid = $field_storage_definition->getTargetEntityTypeId() . '.' . $field_storage_definition->getName();

    if (isset($this->columnInfo[$fsid])) {
      return $this->columnInfo[$fsid];
    }

    return $this->columnInfo[$fsid] = $field_storage_definition->getColumns();
  }

  /**
   * Get a field column name.
   */
  public function getColumns(FieldStorageDefinitionInterface $field_storage_definition) {
    $ret = [];
    $columns = $this->getColumnsInfo($field_storage_definition);

    foreach (array_keys($columns) as $column_name) {
      if (!$field_storage_definition->isBaseField() || $field_storage_definition->isMultiple() || count($columns) !== 1) {
        $ret[$column_name] = $field_storage_definition->getName() . '_' . $column_name;
      }
      else {
        $ret[$column_name] = $field_storage_definition->getName();
      }
    }

    return $ret;
  }

  /**
   * Get table name for field.
   */
  public function getTableName(ContentEntityTypeInterface $entity_definition, FieldStorageDefinitionInterface $field_storage_definition, bool $revision_table = FALSE) : ?string {
    if ($this->allowsSharedTableStorage($field_storage_definition)) {
      $columns = $this->getColumnsInfo($field_storage_definition);

      foreach (array_keys($columns) as $name) {
        $column_name = count($columns) === 1 ? $field_storage_definition->getName() : $field_storage_definition->getName() . '__' . $name;

        foreach ([
          $revision_table ? $entity_definition->getRevisionDataTable() : NULL,
          $revision_table ? $entity_definition->getRevisionTable() : NULL,
          $entity_definition->getDataTable(),
          $entity_definition->getBaseTable(),
        ] as $table) {

          if ($table && $this->database->schema()->fieldExists($table, $column_name)) {
            return $table;
          }
        }
      }
    }
    elseif ($this->requiresDedicatedTableStorage($field_storage_definition)) {
      return $entity_definition->id() . ($revision_table ? '_revision__' : '__') . $field_storage_definition->getName();
    }

    return NULL;
  }

  /**
   * Get table names for field.
   */
  public function getTableNames(ContentEntityTypeInterface $entity_definition, FieldStorageDefinitionInterface $field_storage_definition, bool $revision_tables = FALSE) : array {
    $ret = [];

    if ($this->allowsSharedTableStorage($field_storage_definition)) {
      $columns = $this->getColumnsInfo($field_storage_definition);

      foreach (array_keys($columns) as $name) {
        $column_name = count($columns) === 1 ? $field_storage_definition->getName() : $field_storage_definition->getName() . '__' . $name;

        foreach ([
          $revision_tables ? $entity_definition->getRevisionDataTable() : NULL,
          $revision_tables ? $entity_definition->getRevisionTable() : NULL,
          $entity_definition->getDataTable(),
          $entity_definition->getBaseTable(),
        ] as $table) {
          if ($table && $this->database->schema()->fieldExists($table, $column_name)) {
            $ret[] = $table;
          }
        }
      }
    }
    elseif ($this->requiresDedicatedTableStorage($field_storage_definition)) {
      $ret[] = $entity_definition->id() . ($revision_tables ? '_revision__' : '__') . $field_storage_definition->getName();
    }

    return $ret;
  }

  /**
   * Loads entities by primary id or revision id.
   */
  public function load(string $entity_type, array|SelectInterface $ids = NULL, array $field_names = NULL, string $langcode = NULL, bool $use_revision_tables = FALSE) : array | false {
    $entity_definition = $this->entityTypeManager->getDefinition($entity_type);

    if ($entity_definition instanceof ContentEntityTypeInterface) {
      $res = [];

      $use_revision_tables = $use_revision_tables && $entity_definition->isRevisionable();

      $baseish_tables = array_filter([
        $entity_definition->getRevisionDataTable(),
        $entity_definition->getRevisionTable(),
        $entity_definition->getDataTable(),
        $entity_definition->getBaseTable(),
      ]);

      $baseish_tables = array_combine($baseish_tables, $baseish_tables);
      $keys = $entity_definition->getKeys();
      $id_key = $use_revision_tables && $keys['revision'] ? $keys['revision'] : $keys['id'];
      $field_id_key = $use_revision_tables && $keys['revision'] ? 'revision_id' : 'entity_id';
      $langcode_key = $keys['langcode'] ?? NULL;
      $field_storage_definitions = $this->entityFieldManager->getFieldStorageDefinitions($entity_type);
      $queries = [];

      foreach (['id', 'revision', 'bundle'] as $key_name) {
        $key_value = $keys[$key_name] ?? NULL;

        if ($key_value) {
          $table = $this->getTableName($entity_definition, $field_storage_definitions[$key_value], $use_revision_tables);
          $columns = $this->getColumns($field_storage_definitions[$key_value]);

          foreach ($columns as $column_key => $column_name) {
            $queries[$table][$column_name] = (object) [
              'field' => $key_value,
              'column' => $column_key,
            ];
          }
        }
      }

      if (!$field_names) {
        $field_names = array_keys($field_storage_definitions);
      }

      if (is_array($field_names)) {
        foreach ($field_names as $field_name) {
          $table = $this->getTableName($entity_definition, $field_storage_definitions[$field_name], $use_revision_tables);
          $columns = $this->getColumns($field_storage_definitions[$field_name]);

          if ($table && $columns) {
            foreach ($columns as $column_key => $column_name) {
              $queries[$table][$column_name] = (object) [
                'field' => $field_name,
                'column' => $column_key,
              ];
            }
          }
        }
      }

      foreach ($queries as $table => $query_info) {
        $fields = array_combine(array_keys($query_info), array_keys($query_info));
        $is_shared = !is_null($baseish_tables[$table] ?? NULL);
        $query = NULL;

        if ($is_shared) {
          foreach (['id', 'revision'] as $key_name) {
            $fields[$keys[$key_name]] = $keys[$key_name];
          }

          $query = $this->database->select($table, 't')->fields('t', $fields);

          if ($ids) {
            $query->condition($id_key, $ids, 'IN');
          }

          if ($langcode_key) {
            $query->condition($langcode_key, $langcode ?? 'en');
          }
        }
        else {
          foreach (['entity_id', 'revision_id', 'delta'] as $column_name) {
            $fields[$column_name] = $column_name;
          }

          $query = $this->database->select($table, 't')->fields('t', $fields);

          if ($ids) {
            $query->condition($field_id_key, $ids, 'IN');
          }

          if ($langcode_key) {
            $query->condition('langcode', $langcode ?? 'en');
          }
        }

        if ($query) {
          foreach ($query->execute() as $row) {
            foreach ($row as $column => $value) {
              if ($query_info[$column] ?? NULL) {
                $field_name = $query_info[$column]->field;
                $field_storage_definition = $field_storage_definitions[$field_name];
                $columns = $this->getColumnsInfo($field_storage_definition);
                $column_name = $query_info[$column]->column;
                $column_type = $columns[$column_name]['type'] ?? NULL;
                $row_key = $is_shared ? $id_key : $field_id_key;
                $single_delta = !$field_storage_definition->isMultiple();
                $single_column = (count($columns) === 1);

                if (!is_null($value)) {
                  if ($column_type === 'int') {
                    $value = intval($value);
                  }
                  elseif ($column_type === 'float' || $column_type === 'numeric') {
                    $value = floatval($value);
                  }
                }

                $res[$row->{$row_key}] ??= (object) [];

                if ($single_delta) {
                  if ($single_column) {
                    $res[$row->{$row_key}]->{$field_name} = $value;
                  }
                  else {
                    $res[$row->{$row_key}]->{$field_name} ??= (object) [];
                    $res[$row->{$row_key}]->{$field_name}->{$column_name} = $value;
                  }
                }
                else {
                  $res[$row->{$row_key}]->{$field_name} ??= [];

                  if ($single_column) {
                    $res[$row->{$row_key}]->{$field_name}[$row->delta] = $value;
                  }
                  else {
                    $res[$row->{$row_key}]->{$field_name}[$row->delta] ??= (object) [];
                    $res[$row->{$row_key}]->{$field_name}[$row->delta]->{$column_name} = $value;
                  }
                }
              }
            }
          }
        }
      }

      return $res;
    }

    return FALSE;
  }

}
